#!/usr/bin/env python
# -*- coding: utf-8 -*-

## @package dbus_vrm

# Python
import Queue
import gobject
import logging
import os
import sqlite3
import time
import threading

# Victron imports
from vrmhttp import VrmHTTP, VrmCommandType, vrm_encode
from ve_utils import exit_on_error

logger = logging.getLogger(__name__)

class LOGMODE:
	LoggerDisabled, LogToVRM, LogToStorage = range(3)

QUEUE_SIZE = 4
"""
job = {
	'id': 1231                     # optional, item will only be present when coming from storage
	'data': 'd=1&IMEI=asdfsdfetc'  # already encoded data
	'timestamp': 234234234         # timestamp, used to calculate TO (time-offset)
"""

class HttpEndpoint(object):
	# Threading notes
	# The callback update_last_contact, as well as all operations on buffer are called
	# on the mainthread. See _worker_inner() and _process_results() about the queuing of
	# these operations into the mainthread.
	def __init__(self, logurl, interval, buffer, update_last_contact):
		self._send_queue = Queue.Queue(maxsize=QUEUE_SIZE)
		self._results_queue = Queue.Queue(maxsize=0)
		self._ids_in_queue = []
		self._interval = interval
		self._buffer = buffer
		self._update_last_contact = update_last_contact
		self._feed_from_buffer_to_queue()
		self._worker_thread = threading.Thread(name='HTTPThread', target=self._worker, args=(
			logurl, interval))
		self._request_stop = threading.Event()
		self._process_results_pending = threading.Event()
		self._worker_thread.daemon = True
		self._worker_thread.start()
		logger.info("HttpEndpoint initialised, http worker thread running")

	def send(self, commandtype, data):
		data = vrm_encode(commandtype, data, self._interval)

		if self._buffer.count > 0:
			self._buffer.put(data, int(time.time()))
		else:
			try:
				self._send_queue.put_nowait({'data': data})
			except Queue.Full:
				self._buffer.put(data, int(time.time()))

	def stop(self):
		self._request_stop.set()
		logger.debug("waiting for http worker thread to stop")
		self._worker_thread.join()
		logger.info("HttpEndpoint stopping, http worker thread stopped")

	def _feed_from_buffer_to_queue(self):
		jobs = self._buffer.get_multiple(limit=QUEUE_SIZE, ids_to_skip=self._ids_in_queue)
		try:
			for job in jobs:
				self._send_queue.put_nowait(job)
				self._ids_in_queue.append (job['id'])
		except Queue.Full:
			pass

	def _process_results(self):
		self._process_results_pending.clear()

		# empty the queue into results
		results = []
		try:
			while True:
				results.append(self._results_queue.get_nowait())
				self._results_queue.task_done()
		except Queue.Empty:
			pass

		# walk through the results
		ids_to_remove = []
		last_contact_updated = False
		for (result, job) in results:
			if not result:
				assert 'id' not in job
				self._buffer.put(job['data'], job['timestamp'])
			else:
				if not last_contact_updated:
					self._update_last_contact(int(time.time()))
					last_contact_updated = True
				if 'id' in job:
					ids_to_remove.append(job['id'])
					self._ids_in_queue.remove(job['id'])

		self._buffer.remove_multiple(ids_to_remove)
		self._feed_from_buffer_to_queue()

	def _worker(self, logurl, interval):
		# use try catch to ensure program termination on error
		try:
			self._worker_inner(logurl, interval)
		except Exception as e:
			import traceback
			traceback.print_exc()
			os._exit(1)

	def _worker_inner(self, logurl, interval):
		vrmhttp = VrmHTTP(logurl, interval)
		job = None
		while True:
			if self._request_stop.is_set():
				logger.debug("http worker thread is stopping, since someone asked for it")
				break

			if job is None:
				try:
					job = self._send_queue.get(block=True, timeout=2)
				except Queue.Empty:
					continue

			if not 'timestamp' in job:
				job['timestamp'] = int(time.time())

			result = vrmhttp.send_to_vrm(data=job['data'], timestamp=job['timestamp'])
			logger.debug('result: %s, id: %s' % (result, job.get('id')))

			if not result and 'id' in job:
				continue  # failed, and already in buffer, retry the same job immediately

			self._send_queue.task_done()
			self._results_queue.put((result, job))

			if not self._process_results_pending.is_set():
				self._process_results_pending.set()
				gobject.idle_add(exit_on_error, self._process_results)

			job = None
