#!/usr/bin/python -u
# -*- coding: utf-8 -*-

"""
GENERAL PRINCIPLES ON kWh deltas

Requirements:
For the 'Solar yield' and 'Consumption' tabs on the VRM Portal, we want to get the amount of energy
generated and used. It needs to be split up in the following categories:
- Energy from grid to consumers
- Energy from grid to battery
- Energy from genset to consumers
- Energy from genset to battery
- Energy from PV to consumers
- Energy from PV to battery
- Energy from PV to grid
- Energy from Battery to consumers
- Energy from Battery to grid

When the hour passes, the amount of energy used cq. generated in that hour needs to be sent to the
VRM Database. And during the actual hour, the amount of energy since the start of the hour is
required, imagine looking at the graph and pressing F5 all the time: you will slowly see the bars
growing during that hour.

Input of this process:
- The kWh counters (= energy) of all relevant devices (Multi, BMV, PV Inverters and Solar chargers).
  Note that some of these do not store there counters in flash. So every time, for example, a Multi
  starts up, it starts counting from 0.
- And later, when adding DC consumption or hub-3, it is possibly also necessary also look at the
  direction of DC or AC currents. See hfstore code.

MEASUREMENT INACCURACIES
Assume, for example, you have 3 counters, and since how they are connected electrically, they
relate like a = b + c. But in reality, since all those measurements are not 100% accurate, this
sum will never be 100%.

In below calculations we face the same issue. The chosen approach is to prefer the counters that are
also visible to the user elsewhere, for example to solar yield from the MPPT, or the kWh from the BMV.
The counters in the Multi are not visible to the user.


CALCULATION STATUS:
This code calculates for dc-coupled and ac-coupled systems. It looks at AC consumption for consumption,
not yet at DC consumption.
TODO: add pv on grid (hub-3), and pv on genset to the calculations
TODO: add paralled storage (hub-4) to the calculations
TODO: add possible DC consumption into the calculations


USAGE:
vrmlogger makes an instance of this class. And calls getdeltas() everytime it sends data to the vrm-portal.
Exactly on the hour, vrmlogger calls the same function, including a request to refresh the baseline.


IMPLEMENTATION:
This class, KwhDeltas, uses the DbusDeltas class to store the baselines and calculate the raw deltas. Those
raw deltas are then converted into the required parameters by KwhDeltas.getdeltas(...)
"""

from dbus.mainloop.glib import DBusGMainLoop
import gobject
import argparse
import logging
import os
import sys
import time
import pprint

# Victron imports
sys.path.insert(1, os.path.join(os.path.dirname(__file__), './ext/velib_python'))
from dbusmonitor import DbusMonitor
from dbusdeltas import DbusDeltas

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

class KwhDeltas:
    def __init__(self):
        self._store = {
            'vebus': {
                'services': [], 'class': 'com.victronenergy.vebus',
                'paths': [
                    '/Energy/AcIn1ToInverter',
                    '/Energy/AcIn2ToInverter',
                    '/Energy/AcIn1ToAcOut',
                    '/Energy/AcIn2ToAcOut',
                    '/Energy/InverterToAcIn1',
                    '/Energy/InverterToAcIn2',
                    '/Energy/AcOutToAcIn1',
                    '/Energy/AcOutToAcIn2',
                    '/Energy/InverterToAcOut',
                    '/Energy/OutToInverter']},
            'pvac.genset': {
                'services': [], 'class': 'com.victronenergy.pvinverter',
                'paths': [
                    '/Ac/Energy/Forward',
                    '/Position']},
            'pvac.output': {
                'services': [], 'class': 'com.victronenergy.pvinverter',
                'paths': [
                    '/Ac/Energy/Forward',
                    '/Position']},
            'pvac.grid': {
                'services': [], 'class': 'com.victronenergy.pvinverter',
                'paths': [
                    '/Ac/Energy/Forward',
                    '/Position']},
            'battery': {
                'services': [], 'class': 'com.victronenergy.battery',
                'paths': [
                    '/History/DischargedEnergy',
                    '/History/ChargedEnergy']},
            'solarcharger': {
                'services': [], 'class': 'com.victronenergy.solarcharger',
                'paths': [
                    '/Yield/User']},
            'grid': {
                'services': [], 'class': 'com.victronenergy.grid',
                'paths': [
                    '/Ac/Energy/Forward',
                    '/Ac/Energy/Reverse']},
            'genset': {
                'services': [], 'class': 'com.victronenergy.genset',
                'paths': [
                    '/Ac/Energy/Forward']}}

        """
        # Counters that need to be requested on a high frequency, to distribute the energy
        # into a set counters. Distribution is based on other parameters such as actual
        # battery current.
        self._hfstore = {
            'solarcharger': {
                'services': [], 'class': 'com.victronenergy.solarcharger',
                'paths': [
                    '/Yield/User']}}
        """

    """
    def _init_hfoutput(self):
        self._hfoutput = {
            'SolarchargerToVebus': 0,
            'SolarchargerToBattery': 0}
    """

    def start(self, dbusmonitor, settings_device_gen=None):
        self._dbusmonitor = dbusmonitor

        logger.debug(
            "KwhDeltas.start(): starting up, with self._store: \n%s\n" % pprint.pformat(self._store))

        self._dbusdeltas = DbusDeltas(self._dbusmonitor, self._store, settings_device_gen)
        # self._hfdbusdeltas = DbusDeltas(self._dbusmonitor, self._hfstore)

        # init all instances already existing on the dbus
        services = self._dbusmonitor.get_service_list()
        for name, instance in services.iteritems():
            self.handle_new_service(name, instance)


        # Start timer running hfcalculations
        # self._init_hfoutput()
        # gobject.timeout_add(1000, self._do_hf_calculations)

    def get_dbusmonitortree(self):
        # Get all unique serviceclasses from self._store, and all paths.
        dbusmonitoritems = {}
        # dummy data since DbusMonitor wants it:
        dummy = {'code': None, 'whenToLog': None, 'accessLevel': None}
        for d in self._store.values():
            if d['class'] not in dbusmonitoritems:
                dbusmonitoritems[d['class']] = {}

            for path in d['paths']:
                dbusmonitoritems[d['class']][path] = dummy

        """
        # no longer necessary, because self._hfstore is not necessary for dc-coupled solar power.
        for d in self._hfstore.values():
            if d['class'] not in dbusmonitoritems:
                dbusmonitoritems[d['class']] = {}

            for path in d['paths']:
                dbusmonitoritems[d['class']][path] = dummy

        dbusmonitoritems['com.victronenergy.battery']['/Dc/0/I'] = dummy  # Used for _hfstore processing
        dbusmonitoritems['com.victronenergy.vebus']['/Dc/I'] dummy  # Used for _hfstore processing
        """

        return dbusmonitoritems

    def handle_new_service(self, servicename, instance):
        logger.debug("handle_new_service: %s" % servicename)

        serviceclass = servicename.split('.')
        if len(serviceclass) < 3 or serviceclass[1] != 'victronenergy':
            return

        serviceclass = serviceclass[2]
        if serviceclass == 'vebus':
            # Use all found vebus devices.
            self._dbusdeltas.add_service(servicename, 'vebus')

        elif serviceclass == 'pvinverter':
            # Split pv inverters in three groups: grid, genset and mains
            position = self._dbusmonitor.get_value(servicename, '/Position')
            pvtype = {0: 'pvac.grid', 1: 'pvac.output', 2: 'pvac.genset'}
            logger.debug("%s, p:%s = %s" % (servicename, position, pvtype[position]))
            self._dbusdeltas.add_service(servicename, pvtype[position])
            # TODO: in self._store there is no pvac.genset and pvac.grid yet, add them during init

        elif serviceclass == 'solarcharger':
            # Group all solarchargers together
            # self._hfstore['solarcharger']['services'].append(servicename)
            self._dbusdeltas.add_service(servicename, 'solarcharger')

        elif serviceclass == 'battery':
            # Use the last battery monitor
            # TODO: don't use the last one, use the main battery (needs to be added as a setting)
            self._dbusdeltas.remove_all_services('battery')
            self._dbusdeltas.add_service(servicename, 'battery')
            
        elif serviceclass == 'grid':
            self._dbusdeltas.add_service(servicename, 'grid')

        elif serviceclass == 'genset':
            self._dbusdeltas.add_service(servicename, 'genset')

    def handle_removed_service(self, servicename, instance):
        self._dbusdeltas.remove_service(servicename)

    """
    ==========
    THIS WHOLE SECTION OF THE _hfstore IS DISABLED SINCE AFTER SOME BETTER THINKING, I DIDN'T NEED IT.
    PROBABLY IT WILL BE NECESSARY LATER, FOR HUB-3 AND/OR WHEN DC-CONSUMPTION IS ADDED INTO THE CALCU-
    LATIONS. THEREFORE I LEFT THE CODE IN HERE. SEE ALSO ALL OTHER OUTCOMMENTED LINES OF CODE THAT
    RELATE TO self._hfstore
    ==========

    Below function distributes the solar energy generated since the last call over the two buckets:
    - SolarchargerToVebus
    - SolarchargerToBattery

    The distribution is based on the actual dc-current readings from the VE.Bus system and the battery.
    By calling this function on a relatively high frequency, a quite good result is distracted out of
    the limited data available.

    To determine if distributing the solar power just between vebus and battery will be enough, assume
    the following situation:
    For the first 15 minutes, there is 1kWh solar, and no load, so 1kWh is fed back to the grid. Then,
    for the remaining 45 minutes, there is 2kW solar, 3kWh load, and also 1kWh delivered from the battery
    into VE.Bus. So in total during that hour:
        - 1kWh from the battery to VE.Bus-DC (of which the whole 1kWh went to the load)
        - 3kWh solar to VE.Bus (of which 1kWh to the grid, and 2kWh to the load)

    And from VE.Bus-DC:
        - 1kWh from DC to the grid (must come from Solar, since we don't feed from the battery to the
          grid.)
        - 3kWh from DC to the loads.

    Which can be reconstructed:
        pvdc_to_grid = min(pvdc, vebusdc-grid)
        pvdc_to_loads = pvdc - pvdc_to_grid
        battery_to_loads = bmv_kwh_discharged

    Functions is called periodically from the timer, started from KwhDeltas.start()
    """
    """
    def _do_hf_calculations(self):
        # TODO: add DC consumption to all this.

        if len(self._store['battery']['services']) == 1:
            bmvcurrent = self._dbusmonitor.get_value(self._store['battery']['services'][0], '/Dc/0/I')
        else:
            bmvcurrent = 0
            # TODO (?): raise alert incase there is an MPPT, since we cannot perform kWh calculations for
            # a DC-coupled PV system without BMV.
        if len(self._store['vebus']['services']) == 1:
            vebuscurrent = self._dbusmonitor.get_value(self._store['vebus']['services'][0], '/Dc/I')
            # When > 0, power is going from VE.Bus product into battery. Below 0, other way around.
        else:
            vebuscurrent = 0

        # get the differences
        intervalstart, deltas = self._hfdbusdeltas.get_deltas(refreshbaseline=True)

        pvdc = deltas['solarcharger']['/Yield/User']

        # Add the generated solar power (DC-coupled, so via an MPPT) to the right bucket.
        #
        # When VE.Bus is charging the battery, you know that all MPPT power is going into the
        # battery (or to dc-loads, but we don't take them into account yet). And when VE.Bus
        # is discharging, at least some solar power is going to VE.Bus. Or maybe some to VE.Bus,
        # and some to the battery. That is ofcourse also possible. And that we don't support yet,
        # since we just check if the battery is being charged, and when it is, we attribute all
        # solar energy of that second to the battery. Another approach would be to look at both
        # the current to the battery and the current to VE.Bus, if both positive, split the solar
        # energy over both buckets.
        if vebuscurrent > 0:
            # Multi is charging the battery, so all MPPT energy went into battery as well.
            self._hfoutput['SolarchargerToBattery'] += pvdc
        elif vebuscurrent < 0 and bmvcurrent > 0:
            # Multi is taking power from dc, and battery is still being charged: split MPPT energy
            totalcurrent = float((-1 * vebuscurrent) + bmvcurrent)
            self._hfoutput['SolarchargerToVebus'] += pvdc * -1 * (vebuscurrent / totalcurrent)
            self._hfoutput['SolarchargerToBattery'] += (pvdc - self._hfoutput['SolarchargerToVebus'])
        else:
            # Multi is taking power from dc, and battery is discharging: all MPPT energy to multi
            self._hfoutput['SolarchargerToVebus'] += pvdc

        return True  # keep timer running
    """

    # Core of this class. Performs calculations and returns all kwhdeltas in a dict which is formatted to
    # be sent straight to VRM.
    def getdeltas(self, refreshbaseline):

        baselinetimestamp, deltas, servicecounts = self._dbusdeltas.get_deltas(refreshbaseline)

        logger.debug("===== raw deltas start =====")
        logger.debug(pprint.pformat(deltas))
        logger.debug("====== raw deltas end ======")

        pvac_output = deltas['pvac.output']['/Ac/Energy/Forward']

        vebus = deltas['vebus']

        pvac_to_battery = min(vebus['/Energy/AcOutToDc'], pvac_output)
        pvac_to_consumers = max(pvac_output - vebus['/Energy/AcOutToGrid'] - vebus['/Energy/AcOutToDc'], 0)
        pvac_to_grid = min(vebus['/Energy/AcOutToGrid'], pvac_output)

        logger.debug("pvac_to_grid, battery, consumers: %s, %s, %s" % (pvac_to_grid, pvac_to_battery, pvac_to_consumers))
        logger.debug("sum of the three: %s" % (pvac_to_grid + pvac_to_consumers + pvac_to_battery))

        """
        Distribution of DC-Coupled solar power over the battery, consumers and grid:

        Just looking at the dc-coupled solar power at the end of the hour can be good enough. Take following
        example:
        At the end of the hour you have 2kW of generated PV (dc), 2kW of energy that went into the battery, and
        the Multi tells you it charged 1kW. Calculation: From the 2kW into the battery, 1kW came from the Multi.
        the rest has to come from dc-pv. The remainder of the PV went into the Multi. Any energy that was fed back
        to the grid came from the PV, since the Multi doesn't feed energy from the battery into the grid. Any
        remaining dc-pv after that went to the consumers
        """
        totalvebustobattery = vebus['/Energy/GridToDc'] + vebus['/Energy/GensetToDc'] + vebus['/Energy/AcOutToDc']
        pvdc_to_grid = min(vebus['/Energy/DcToGrid'], deltas['solarcharger']['/Yield/User'])
        pvdc_to_battery = max(deltas['battery']['/History/ChargedEnergy'] - totalvebustobattery, 0)
        # Maximize the pvdc_to_battery to the power available from solar:
        pvdc_to_battery = min(deltas['solarcharger']['/Yield/User'] - pvdc_to_grid, pvdc_to_battery)
        pvdc_to_consumers = max(deltas['solarcharger']['/Yield/User'] - pvdc_to_battery - pvdc_to_grid, 0)

        """
        # Option 2: use the output of the hfstore
        pvdc_to_battery = self._hfoutput['SolarchargerToBattery']
        pvdc_to_grid = min(vebus['/Energy/DcToGrid'], self._hfoutput['SolarchargerToVebus'])
        pvdc_to_consumers = self._hfoutput['SolarchargerToVebus'] - pvdc_to_grid
        """

        logger.debug("pvdc to grid, battery, consumers: %s, %s, %s" % (pvdc_to_grid, pvdc_to_battery, pvdc_to_consumers))
        logger.debug("sum of the three: %s" % (pvdc_to_grid + pvdc_to_consumers + pvdc_to_battery))

        # Reset energy counters that keep the result of the hf calculations
        # if refreshbaseline:
        #     self._init_hfoutput()

        # Grid meter related data and PV inverters between grid meter and multi
        pvin_to_battery, pvin_to_grid, pvin_to_consumers, grid_to_battery, grid_to_consumers = \
            self._get_ac_in_values(vebus, deltas, servicecounts, 'Grid')

        # Generator related data.
        pvin_genset_to_battery, pvin_genset_to_grid, pvin_genset_to_consumers, genset_to_battery, genset_to_consumers = \
            self._get_ac_in_values(vebus, deltas, servicecounts, 'Genset')
        assert pvin_genset_to_grid == 0

        result = {}
        result['Gc'] = grid_to_consumers
        result['Gb'] = grid_to_battery
        result['gc'] = genset_to_consumers
        result['gb'] = genset_to_battery
        result['Pb'] = pvac_to_battery + pvdc_to_battery + pvin_to_battery + pvin_genset_to_battery # PV to battery
        result['Pc'] = pvac_to_consumers + pvdc_to_consumers + pvin_to_consumers + pvin_genset_to_consumers # PV to consumers
        result['Pg'] = pvac_to_grid + pvdc_to_grid + pvin_to_grid # PV to grid

        if servicecounts['battery']['/History/DischargedEnergy'] > 0:
            # There is a battery monitor with kWh counting installed, use it. We assume all power
            # from the battery went to the loads, since the Multi does not feed power from the battery
            # back into the grid.
            logger.debug("Using BMV data for Battery-to-consumers")
            result['Bc'] = deltas['battery']['/History/DischargedEnergy']
            result['Bg'] = 0
        else:
            # no battery monitor, so just use the counters from VE.Bus
            logger.debug("Using VE.Bus data for Battery-to-consumers")
            result['Bc'] = vebus['/Energy/DcToAcOut'] # Battery to consumers
            result['Bg'] = vebus['/Energy/DcToGrid'] # Battery to grid, dummy value: should be 0

        result['Et'] = time.strftime("%Y%m%d%H", time.gmtime(baselinetimestamp))

        logger.debug("Calculated deltas, with baseline-timestamp: %s, age (m): %s" % (
            baselinetimestamp,
            (time.time() - baselinetimestamp) / 60))
        return result

    def _get_ac_in_values(self, vebus, deltas, servicecounts, ac_in_source):
        if servicecounts[ac_in_source.lower()]['/Ac/Energy/Forward'] == 0:
            return self._get_ac_in_values_no_gridmeter(vebus, deltas, ac_in_source)
        grid_in = deltas[ac_in_source.lower()]['/Ac/Energy/Forward']
        grid_out = deltas[ac_in_source.lower()].get('/Ac/Energy/Reverse', 0)
        pvac_in = deltas['pvac.' + ac_in_source.lower()]['/Ac/Energy/Forward']
        acin_to_battery = vebus['/Energy/' + ac_in_source + 'ToDc']
        acin_to_acout = vebus['/Energy/' + ac_in_source + 'ToAcOut']

        pvin_to_battery = min(pvac_in, acin_to_battery)
        pvac_in -= pvin_to_battery
        acin_to_battery -= pvin_to_battery

        pvin_to_grid = min(grid_out, pvac_in)
        grid_out -= pvin_to_grid
        pvac_in -= pvin_to_grid

        pvin_to_ac_out = min(pvac_in, acin_to_acout)
        pvac_in -= pvin_to_ac_out
        acin_to_acout -= pvin_to_ac_out

        grid_to_battery = min(grid_in, acin_to_battery)
        grid_in -= grid_to_battery
        acin_to_battery -= grid_to_battery

        grid_to_ac_out = min(grid_in, acin_to_acout)
        grid_in -= grid_to_ac_out
        acin_to_acout -= grid_to_ac_out

        pvin_to_consumers = pvac_in + pvin_to_ac_out
        grid_to_consumers = grid_in + grid_to_ac_out

        return pvin_to_battery, pvin_to_grid, pvin_to_consumers, grid_to_battery, grid_to_consumers


    def _get_ac_in_values_no_gridmeter(self, vebus, deltas, ac_in_source):
        pvac_in = deltas['pvac.' + ac_in_source.lower()]['/Ac/Energy/Forward']
        acin_to_battery = vebus['/Energy/' + ac_in_source + 'ToDc']
        acin_to_acout = vebus['/Energy/' + ac_in_source + 'ToAcOut']

        pvin_to_battery = min(pvac_in, acin_to_battery)
        pvac_in -= pvin_to_battery
        acin_to_battery -= pvin_to_battery

        pvin_to_ac_out = min(pvac_in, acin_to_acout)
        pvac_in -= pvin_to_ac_out
        acin_to_acout -= pvin_to_ac_out

        #      pvin_to_battery, pvin_to_grid, pvin_to_consumers, grid_to_battery, grid_to_consumers
        return pvin_to_battery, pvac_in, pvin_to_ac_out, acin_to_battery, acin_to_acout


# === All code below is to debug this class without needing to run the the full vrmlogger ===
kwhdeltas = None

def handle_timer():
    try:
        deltas = kwhdeltas.getdeltas(refreshbaseline=False)
    
        logger.debug("Final deltas: \n%s" % pprint.pformat(deltas))
    
        return True
    except:
        import traceback 
        traceback.print_exc() 
        sys.exit(1) 
  

# Called from the dbusmonitor
def handle_new_service(servicename, deviceinstance):
    kwhdeltas.handle_new_service(servicename, deviceinstance)


# Called from the dbusmonitor
def handle_removed_service(servicename, deviceinstance):
    kwhdeltas.handle_removed_service(servicename, deviceinstance)


def main():
    global kwhdeltas

    logging.basicConfig(level=logging.DEBUG)

    # Have a mainloop, so we can send/receive asynchronous calls to and from dbus
    DBusGMainLoop(set_as_default=True)

    kwhdeltas = KwhDeltas()
    kwhdeltas.start(
        DbusMonitor(
            dbusTree=kwhdeltas.get_dbusmonitortree(),
            deviceAddedCallback=handle_new_service,
            deviceRemovedCallback=handle_removed_service)
    )

    gobject.timeout_add(3000, handle_timer)

    # Start and run the mainloop
    logging.info("Starting mainloop, responding on only events from now on. Press ctrl-Z to see the deltas")
    mainloop = gobject.MainLoop()
    mainloop.run()


if __name__ == "__main__":
    main()
