#!/usr/bin/python -u
# -*- coding: utf-8 -*-

import logging
import os
import gobject
from sys import path as syspath
from time import sleep

# Victron imports
syspath.insert(1, os.path.join(os.path.dirname(__file__), './ext/velib_python'))

"""
NetworkTrafficDeltas

Known limitations:
    - when an interface dissapears (someone removes a wifi stick), its traffic since the last baseline-
      refresh will be lost. This will mean that if you remove a stick on 9:59 (worst case), all traffic
      between 9:00 and 9:59 will be lost.
    - it will check for new interfaces every minute: traffic between insertion and the end of that
      minute is lost.
"""

logger = logging.getLogger(__name__)

class NetworkTrafficDeltas:
    def __init__(self):
        self._baseline = {}
        self.getdeltas(True)
        gobject.timeout_add(60 * 1000, self._check_for_new_interfaces)

    def _check_for_new_interfaces(self):
        try:
            # by calling getdeltas, without refreshing the baseline, any new interfaces will automatically be
            # added to the baseline.
            self.getdeltas(refreshbaseline=False)
            return True  # keep timer running
        except:
            import traceback
            import sys
            traceback.print_exc() 
            sys.exit(1) 
            
    # Core of this class. Performs calculations and returns all kwhdeltas in a dict which is formatted to
    # be sent straight to VRM.
    def getdeltas(self, refreshbaseline):
        newcounters = {}
        for root, dirs, files in os.walk('/sys/class/net/', topdown=False):
            for interface in dirs:
                newcounters[interface] = {
                    'tx': self._readcounter(os.path.join(root, interface, './statistics/tx_bytes')),
                    'rx': self._readcounter(os.path.join(root, interface, './statistics/rx_bytes'))}

        # Newcounters example: {'eth0': {'rx': 5009419, 'tx': 1099829}, 'lo': {'rx': 103981, 'tx': 103981}}
        result = {'tx': 0, 'rx': 0}
        for interface, base in self._baseline.iteritems():
            if interface not in newcounters:
                continue

            if interface.startswith('eth') or interface.startswith('wlan'):
                result['tx'] += max(newcounters[interface]['tx'] - base['tx'], 0)
                result['rx'] += max(newcounters[interface]['rx'] - base['rx'], 0)

        if refreshbaseline:
            self._baseline = newcounters
        else:
            # add interfaces that are new, and therefore not in the baseline yet. This would happen if
            # someone plugs in a wifistick
            for interface, counters in newcounters.iteritems():
                if interface in self._baseline:
                    continue

                logger.info("New network interface found: %s. Adding its counters to baseline " % interface)
                self._baseline[interface] = {'tx': counters['tx'], 'rx': counters['rx']}

        # Example of result: {'rx': 8944, 'tx': 12495}
        return result

    def _readcounter(self, path):
        try:
            fileObject = open(path)
            result = int(fileObject.readline())
            fileObject.close()
        except Exception, ex:
            logger.error('_readcounters(): exception reading %s: %s.' % (path, ex))
            result = 0

        return result


# === All code below is to debug this class without needing to run the the full vrmlogger ===
trafficdeltas = None

def main():
    global trafficdeltas

    logging.basicConfig(level=logging.DEBUG)

    trafficdeltas = NetworkTrafficDeltas()

    # Start and run the mainloop
    logger.info("Starting mainloop, responding on only events")
    mainloop = gobject.MainLoop()
    mainloop.run()


if __name__ == "__main__":
    main()
