#!/usr/bin/env python
# -*- coding: utf-8 -*-

## @package dbus_vrm

# Python
import logging
import sqlite3

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

class QueueError(Exception):
	def __init__(self, message):
		Exception.__init__(self, message)
		self.message = message

	def __repr__(self):
		return self.message


class NonVolatileQueue(object):
	# max_age         the maximum age of items in the queue, in seconds. Everytime when adding an
	#                 item, all items with age > maxage will be deleted.
	# update_count    function that will be called every time the number of items in store changes.
	#                 the function passed must have one parameter.
	def __init__(self, max_age, path, update_count=None, update_oldest_timestamp=None):
		self._max_age = max_age
		self.path = path
		self._update_count = update_count
		self._update_oldest_timestamp = update_oldest_timestamp
		self.open()

	def open(self):
		def dict_factory(cursor, row):
			d = {}
			for idx, col in enumerate(cursor.description):
				d[col[0]] = row[idx]
			return d

		try:
			self._sqlite_dbconn = sqlite3.connect(self.path)
			self._sqlite_dbconn.row_factory = dict_factory
			cursor = self._sqlite_dbconn.cursor()
			cursor.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='queue';")
			if not cursor.fetchall():
				# Setting auto_vacuum to zero, from the documentation:
				#    Unused database file pages are added to a "freelist" and
				#    reused for subsequent inserts. So no database file space is
				#    lost.  However, the database file does not shrink.
				# Which is great for flash usage.
				cursor.execute('PRAGMA auto_vacuum = 0;')
				cursor.execute("CREATE TABLE queue (id integer primary key autoincrement, timestamp, data)")
				self._sqlite_dbconn.commit()
				logger.info("db %s did not exist yet. Created it" % self.path)

				self.oldest_item = None
				if self._update_oldest_timestamp:
					self._update_oldest_timestamp(None)

				self.count = 0
			else:
				self._refresh_oldest_item()
				self._refresh_count()
			# Create an index on the timestamp column, because some of the queries used in this class select
			# ranges based on that column. This will improve performance a lot when the database gets bigger.
			# On a database with 30000+ entries cpu usage of vrmlogger dropped from 75% to 5%.
			cursor.execute("CREATE INDEX IF NOT EXISTS queue_timestamp ON queue(timestamp)")
		except sqlite3.DatabaseError,e:
			if e.message == 'database disk image is malformed':
				raise QueueError('queue is corrupt')

	# add a new item
	def put(self, data, timestamp):
		# since sqlite does not do type checking, lets do it ourselves
		assert isinstance(timestamp, int)

		try:
			count = self.count
			cursor = self._sqlite_dbconn.cursor()

			cursor.execute("DELETE FROM queue WHERE timestamp < (strftime('%s', 'now') - ?)", (self._max_age,))
			if cursor.rowcount > 0:
				count -= cursor.rowcount
				logger.debug("Deleted %d rows because they expired. (max age = %d)" % (cursor.rowcount, self._max_age))

			if count == 0:
				logger.info("Starting to put data into storage (%s). Count was 0" % self.path)

			cursor.execute('INSERT INTO queue(timestamp, data) VALUES (?, ?)', (timestamp, data))
			count += 1
			self._sqlite_dbconn.commit()

			self.count = count

			logger.debug("Added item with id %d, new count is %d" % (cursor.lastrowid, self.count))

			self._refresh_oldest_item()
		except sqlite3.OperationalError,e:
			if e.message == 'unable to open database file':
				raise QueueError('queue not found')
			elif e.message == 'disk I/O error':
				raise QueueError('I/O error')
			else:
				raise

	@property
	def count(self):
		return self._count

	@count.setter
	def count(self, count):
		self._count = count
		if self._update_count:
			self._update_count(count)

	def remove(self, id):
		self.remove_multiple([id])

	def remove_multiple(self, ids):
		if not ids:
			return

		cursor = self._sqlite_dbconn.cursor()
		cursor.execute('DELETE FROM queue WHERE id IN (%s)' % ','.join('?' * len(ids)), ids)
		assert cursor.rowcount == len(ids), \
			"Buffer %s, expected to delete ids %s, but delete rowcount (%d) was inconsistent!" % \
			(self.path, ids, cursor.rowcount)
		self._sqlite_dbconn.commit()
		logger.debug("Removed ids %s" % ids)
		self.count -= len(ids)
		self._refresh_oldest_item()
		if self.oldest_item is None:
			logger.info("No more items in %s" % self.path)

	def get(self, ids_to_skip=[]):
		r = self.get_multiple(limit=1, ids_to_skip=ids_to_skip)
		return r[0] if r else None

	def get_multiple(self, limit=1, ids_to_skip=[]):
		cursor = self._sqlite_dbconn.cursor()
		a = []
		a.extend(ids_to_skip)
		a.append(limit)
		cursor.execute(
			'SELECT id, timestamp, data FROM queue WHERE id NOT IN (%s) ORDER BY timestamp LIMIT ?' %
			','.join('?' * len(ids_to_skip)), a)

		return cursor.fetchall()

	def close(self):
		self._sqlite_dbconn.close()

	def _refresh_oldest_item(self):
		self.oldest_item = self.get()
		if self._update_oldest_timestamp:
			self._update_oldest_timestamp(self.oldest_item['timestamp'] if self.oldest_item else None)

	def _refresh_count(self):
		c = self._sqlite_dbconn.cursor()
		c.execute("SELECT COUNT(*) as cnt FROM queue")
		self.count = c.fetchone()['cnt']


if __name__ == "__main__":
	import time
	logging.basicConfig()
	logger.setLevel(logging.DEBUG)
	o = NonVolatileQueue(60 * 60 * 24 * 2, path="./nonvolatilequeue_test.sqlite3")
	print "oldest item: ", o.oldest_item
	o.put('hoi', int(time.time()))
	print 'count: %d' % o.count
	print "oldest item: ", o.oldest_item
	o.remove(o.oldest_item['id'])
	print "oldest item:", o.oldest_item
	print "count: %s" % o.count
	o._refresh_count()
	print "count after doing a hard refresh: %s" % o.count
