#!/usr/bin/env python
# -*- coding: utf-8 -*-

## @package dbus_vrm

# Python
from datetime import datetime
import logging
import os
import sqlite3
import subprocess
import time
import shutil

# Victron imports
from vrmhttp import vrm_encode
from nonvolatilequeue import NonVolatileQueue, QueueError
logger = logging.getLogger(__name__)

# The location where the currently used storage_folder is saved.
# TODO: why not in localsettings?
LAST_KNOWN_STORAGEPATH_FILE = os.path.join(os.path.dirname(__file__), 'last_known_storagepath')

def read_last_used_storage_folder():
	storage_folder = None
	try:
		with open(LAST_KNOWN_STORAGEPATH_FILE) as handle:
			storage_folder = handle.read()
		if not os.path.exists(storage_folder):
			return None

	except IOError:
		return None

	return storage_folder if storage_folder != '' else None


def write_currently_used_storage_folder(folder):
	with open(LAST_KNOWN_STORAGEPATH_FILE, 'w') as handle:
		handle.write(folder)


class StorageEndpoint(object):
	def __init__(self, interval, buffer, dbusconn, update_storage_count=None, update_logfiles_count=None, 
				update_mount_state=None):
		self._interval = interval
		self._update_storage_count_f = update_storage_count
		self._update_logfiles_count_f = update_logfiles_count
		self._update_mount_state_f = update_mount_state
		self._storage = None
		self._buffer = buffer
		self._storage_folder = read_last_used_storage_folder()
		if self._storage_folder:
			self._mount_event('/dev/xxx', self._storage_folder)

		dbusconn.add_signal_receiver(self._mount_event, signal_name='mount')
		dbusconn.add_signal_receiver(self._umount_event, signal_name='umount')

	def send(self, commandtype, data):
		data = vrm_encode(commandtype, data, self._interval)

		if self._storage is None:
			self._buffer.put(data, int(time.time()))
		else:
			assert self._buffer.count == 0, \
				'Expected buffer to be empty, but it is %s' % self._buffer.count
			if self._get_storage_path() != self._storage.path:
				logger.info("New storage path (%s) != existing (%s), switching to the new one"
					% (self._get_storage_path(), self._storage.path))
				self._storage.close()
				self._storage = None
				self._mount_event('/dev/xxx', self._storage_folder)

			self._make_room_on_storage_if_needed()
			self._update_logfiles_count()
			try:
				self._storage.put(data, int(time.time()))
			except QueueError,e:
				if e.message == 'queue not found':
					self._umount_event('/dev/xxx', self._storage_folder)
				elif e.message == 'I/O error':
					pass
				else:
					raise

	def eject(self):
		self._update_mount_state(3)
		f = self._storage_folder
		self._umount_event('/dev/xxx', f)
		r = subprocess.call(['umount', f])
		logging.info('umount {} with result {}'.format(r, f))

	def _mount_event(self, device, mountpoint):
		logger.info('Mount event:{} -> {}'.format(device, mountpoint))
		if self._storage:
			logger.info('We are already logging to external storage. Ignoring the new')
			return;

		if not os.access(mountpoint, os.W_OK):
			logger.warning('Mount point {} is not writeable - ignoring.'.format(mountpoint))
			return

		write_currently_used_storage_folder(mountpoint)
		self._storage_folder = mountpoint
		storage_path = self._get_storage_path()
		logger.info('Changed storage path to: {}'.format(storage_path))
		self._update_mount_state(1)

		# if there is data in the current buffer, move that data to the new location.
		if self._buffer.count > 0:
			self._buffer.close()
			logging.info('moving db from {} to {}'.format(self._buffer.path, storage_path))
			shutil.move(self._buffer.path, storage_path)
			journal_path = storage_path + '-journal'
			if os.path.exists(journal_path):
				logger.debug('{}: old database journal still present'.format(journal_path))
				os.remove(journal_path)
			self._buffer.open()

		for retry in range(2):
			try:
				self._storage = NonVolatileQueue(
					max_age=None,
					path=storage_path,
					update_count=self._update_storage_count_f,
					update_oldest_timestamp=None)
				break
			except QueueError,e:
				if e.message == 'queue is corrupt':
					logger.error('Database error: {}. Will try to move the database aside'.format(e))
					for i in range(50):
						p = storage_path.replace('.sqlite3', '-corrupt{}.sqlite3'.format(i))
						if not os.path.exists(p):
							os.rename(storage_path, p)
							break
				else:
					raise

		self._update_logfiles_count()

	def _umount_event(self, device, mountpoint):
		logger.info('Umount event:{} -> {}'.format(device, mountpoint))
		if self._storage_folder == mountpoint:
			if self._storage is not None:
				self._storage.close()
				self._storage = None
			if self._update_storage_count_f is not None:
				self._update_storage_count_f(None)
			if self._update_logfiles_count_f is not None:
				self._update_logfiles_count_f(None)
			self._update_mount_state(0)
			write_currently_used_storage_folder('')
			self._storage_folder = None

	def _get_storage_path(self):
		return '{0}/vrmdata-{1}.sqlite3'.format(
			self._storage_folder,
			datetime.now().strftime('%Y-%m'))

	def _update_logfiles_count(self):
		c = 0
		for filename in os.listdir(self._storage_folder):
			if filename.startswith('vrmdata-') and filename.endswith('.sqlite3'):
				c += 1
		if self._update_logfiles_count_f != None:
			self._update_logfiles_count_f(c)

	def _update_mount_state(self, state):
		if self._update_mount_state_f != None:
			self._update_mount_state_f(state)

	def _make_room_on_storage_if_needed(self):
		if self._storage_folder is None or self._enough_free_space_on_storage():
			return

		logger.debug("Not enough space left on storage, removing databases")

		existing_databases = []
		for filename in os.listdir(self._storage_folder):
			if filename.startswith('vrmdata-') and filename.endswith('.sqlite3'):
				existing_databases.append(filename)
		existing_databases.sort()  # We know they are written in alphanumerical order

		for oldest_database in existing_databases:
			full_path = os.path.join(self._storage_folder, oldest_database)

			if os.path.normpath(full_path) == os.path.normpath(self._storage.path):
				logger.warning("Deleting {} (current one!)to make room for new data.".format(full_path))
				self._storage.close()
				os.unlink(full_path)
				self._storage.open()
			else:
				logger.warning("Deleting {} to make room for new data.".format(full_path))
				os.unlink(full_path)

			if self._enough_free_space_on_storage():
				break

	def _enough_free_space_on_storage(self):
		stat = os.statvfs(self._storage_folder)
		free_bytes = stat.f_bsize * stat.f_bfree
		logger.debug("There are {} bytes free on storage".format(free_bytes))
		return (free_bytes > 4096)
