# Python
import logging
import os
import shutil
import sqlite3
import sys
import tempfile
import time
import unittest
from urlparse import parse_qs

# Local
from conversions import Conversions
from vrmhttpflash import VrmHTTPFlash, LAST_KNOWN_STORAGEPATH_FILE
from vrmhttp import VrmCommandType


logger = logging.getLogger(__file__)


# Currently we don't mock time, thus we can only compare relative times. This
# is the max delta for when something should have happened 'recently'.
MAX_TIME_DELTA = 1


class ConversionTests(unittest.TestCase):

	def test_c01_instantiated(self):
		obj = Conversions()
		self.assertEqual(1, obj.convert('TTG', 3600))

	def test_c02_uninstantiated(self):
		self.assertEqual(1, Conversions.convert('TTG', 3600))

	def test_c03_alarm(self):
		self.assertEqual(0, Conversions.convert('Alarm', 0))
		self.assertEqual(1, Conversions.convert('Alarm', 1))
		self.assertEqual(1, Conversions.convert('Alarm', 2))

class VrmHttpFlashTests(unittest.TestCase):

	def setUp(self, attribute_name='CV[0]', attribute_start_value=1234.5):
		self.tmp_dir = tempfile.mkdtemp(suffix='vrmlogger_test')
		os.mkdir('tmp_vrm_flash_backlog')
		self.tmp_flash_path = os.path.join('tmp_vrm_flash_backlog', 'test_vrm_flash_backlog.sqlite3')
		self._attribute_name = attribute_name
		self._attribute_value = attribute_start_value
		self._successfully_delivered = []
		self.setupVrmHttpFlash()

	def tearDown(self):
		del self.flash
		shutil.rmtree(self.tmp_dir)
		shutil.rmtree('tmp_vrm_flash_backlog')
		if (os.path.exists(LAST_KNOWN_STORAGEPATH_FILE)):
			os.unlink(LAST_KNOWN_STORAGEPATH_FILE)

	def setupVrmHttpFlash(self):
		self.flash = VrmHTTPFlash(backlogpath=self.tmp_flash_path, logmode=self._logmode, dbusservice={})
		self.flash._send_data_instance = self.undefined_internet_connection

	def break_the_internet(self):
		logger.info('Breaking the internet')
		self.flash._send_data_instance = self.broken_internet_connection

	def fix_the_internet(self):
		logger.info('Fixing the internet')
		self.flash._send_data_instance = self.working_internet_connection

	def get_queue_length(self, sqlite_path):
		if sqlite_path is None or not os.path.exists(sqlite_path):
			return 0  # VrmFlash creates the databases as late as possible
		conn = sqlite3.connect(sqlite_path)
		cursor = conn.cursor()
		cursor.execute("SELECT count(*) FROM queue")
		return cursor.fetchone()[0]

	def send_data_to_vrm(self, value=None):
		if value is None:
			value = self._attribute_value
			self._attribute_value += 1
		data = {self._attribute_name: value}
		self.flash.send_to_vrm(VrmCommandType.SENDDATA, data, blocking=True)
		if self.flash._logmode == 1:
			self.flash._send_data_from_backlog()  # Happens on a timed trigger in the real VrmHttpFlash

	def undefined_internet_connection(self, *args, **kwargs):
		self.fail("You should configure an internet connection for your testcase.")

	def working_internet_connection(self, *args, **kwargs):
		_parsed = parse_qs(args[1])
		_current_value = float(_parsed[self._attribute_name][0])
		if self._successfully_delivered:
			self.assertGreater(_current_value, self._successfully_delivered[-1])
		self._successfully_delivered.append(_current_value)
		self.simple_working_internet_connection(*args, **kwargs)

	def simple_working_internet_connection(self, logurl, data, queue=None, custom_queue_info=None):
		'''Mock version of _send_data_instance - always succeeds'''
		logger.info('Fake working internet connection for {}'.format(data))
		result = True
		if queue is not None:
			queue.put([result, custom_queue_info])
		return result

	def broken_internet_connection(self, logurl, data, queue=None, custom_queue_info=None):
		'''Mock version of _send_data_instance - always fails'''
		logger.info('Fake broken internet connection for {}'.format(data))
		result = False
		if queue is not None:
			queue.put([result, custom_queue_info])
		return result

	def assertBacklogs(self, location=None, get=None, flash=None, storage=None):
		if self.flash._logmode == 1:
			self.flash.process_httpcall_results()  # To process the last result - VrmFlash calls this periodically
		self.assertNotIn(None, [get, flash, storage, location], 'Please write detailed testcases.')
		self.assertEqual(location, self.flash._active_backlog)
		self.assertEqual(self.flash.get_backlog_size(), get)
		self.assertEqual(flash, self.get_queue_length(self.flash._sqlite_flash_path))
		self.assertEqual(storage, self.get_queue_length(self.flash._sqlite_storage_path))

	def assertDeliveries(self, expected):
		'''Assert the number of data entries that was delivered to the VRM portal'''
		self.assertEqual(expected, len(self._successfully_delivered))

class VrmPortalBacklogTests(VrmHttpFlashTests):
	'''Test requirements and scenarios of VrmHttpFlash

	While not the case in real life, for the testcases it is assumed that the
	attribute value passed to the VRM Portal is strictly increasing over time.
	This way the requirement that all data should be logged chronologically can
	be tested by looking at the data attribute.
	'''

	def setUp(self, attribute_name='CV[0]', attribute_start_value=1234.5):
		self._logmode = 1
		VrmHttpFlashTests.setUp(self, attribute_name=attribute_name, attribute_start_value=attribute_start_value)

	def test_p01_empty_on_startup(self):
		self.assertBacklogs(location='', get=0, flash=0, storage=0)

	def test_p02_send_with_working_internet(self):
		self.fix_the_internet()
		self.assertBacklogs(location='', get=0, flash=0, storage=0)
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.assertBacklogs(location='', get=0, flash=0, storage=0)

	def test_p03_inserting_into_backlog(self):
		self.break_the_internet()
		self.assertBacklogs(location='', get=0, flash=0, storage=0)
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.assertBacklogs(location='flash', get=4, flash=4, storage=0)

	def test_p04_backlog_should_try_vrm_portal_first(self):
		'''Test logging with a flapping internet.

		All data should first be offered to the portal. Only when the
		internet is failing should it be kept in backlog. As soon as the
		internet comes back data should go to the portal again, oldest
		items first. After all backlog items have been transmitted it
		should stop using the memory backlog.'''

		self.fix_the_internet()
		self.assertBacklogs(location='', get=0, flash=0, storage=0)

		# Working internet, so backlog should not be used.
		self.send_data_to_vrm(124.5)
		self.send_data_to_vrm(125.5)
		self.send_data_to_vrm(126.5)

		self.assertAlmostEqual(0, self.flash._dbusservice['/Logscript/OldestBacklogItemAge'])
		self.assertBacklogs(location='', get=0, flash=0, storage=0)
		self.assertEqual([124.5, 125.5, 126.5], self._successfully_delivered)

		# Failing internet, so backlog should be used
		self.break_the_internet()
		self.send_data_to_vrm(127.5)
		self.send_data_to_vrm(128.5)

		timestamp_of_backlog_x = self.flash._dbusservice['/Logscript/OldestBacklogItemAge']
		self.assertGreater(MAX_TIME_DELTA, abs(time.time() - timestamp_of_backlog_x))
		self.assertBacklogs(location='flash', get=2, flash=2, storage=0)
		self.assertEqual([124.5, 125.5, 126.5], self._successfully_delivered)

		# Hooray, the internet works again.
		self.fix_the_internet()
		self.flash._send_data_from_backlog()  # Normally triggered by a timer
		self.send_data_to_vrm(129.5)
		self.send_data_to_vrm(130.5)
		self.flash._send_data_from_backlog()  # Normally triggered by a timer
		self.send_data_to_vrm(131.5)

		timestamp_of_backlog_y = self.flash._dbusservice['/Logscript/OldestBacklogItemAge']
		self.assertTrue(MAX_TIME_DELTA, abs(time.time() - timestamp_of_backlog_y))
		# WARNING:
		# - By design, the following checks identity, not equality!
		# - Comparing the identify of integers only works with big numbers (>255). Small numbers
		#   are cached and recycling, giving them the same identity.
		self.assertTrue(timestamp_of_backlog_x is not timestamp_of_backlog_y)  # Checks identity, not equality!
		self.assertBacklogs(location='', get=0, flash=0, storage=0)
		self.assertEqual([124.5, 125.5, 126.5, 127.5, 128.5, 129.5, 130.5, 131.5], self._successfully_delivered)

		# The full backlog has been trasmitted, if we try to send more data the VRM should revert
		# back to the 'no backlog active' state.
		self.flash._send_data_from_backlog()  # Normally triggered by a timer
		self.assertBacklogs(location='', get=0, flash=0, storage=0)

		self.send_data_to_vrm(132.5)
		self.assertEqual(0, self.flash._dbusservice['/Logscript/OldestBacklogItemAge'])
		self.assertBacklogs(location='', get=0, flash=0, storage=0)
		self.assertEqual([124.5, 125.5, 126.5, 127.5, 128.5, 129.5, 130.5, 131.5, 132.5], self._successfully_delivered)

	def test_p05_resetting_vrmlogger_with_flash_backlog(self):
		'''The flash backlog should be restored on startup if it contains data from a previous run.

		Based on https://wiki.victronenergy.com/wiki/rend/ccgx/v1.16/v1.16_beta_backlogging_does_not_work
		'''
		self.break_the_internet()

		self.send_data_to_vrm(0.12)
		self.send_data_to_vrm(1.23)
		self.send_data_to_vrm(2.34)
		self.send_data_to_vrm(3.45)
		self.send_data_to_vrm(4.56)
		self.send_data_to_vrm(5.67)
		self.send_data_to_vrm(6.78)
		self.assertBacklogs(location='flash', get=7, flash=7, storage=0)
		self.assertEqual([], self._successfully_delivered)

		logging.debug('Destroying existing Flash instance...')
		del self.flash

		logging.debug('Creating new Flash instance...')
		self.setupVrmHttpFlash()
		self.assertBacklogs(location='flash', get=7, flash=7, storage=0)
		self.fix_the_internet()

		# The following line just sends from the backlog, but becasue the result is not handled, the backlog count is not
		# yet updated
		self.flash._send_data_from_backlog()  # 0.12, Normally triggered by a timer
		self.flash._send_data_from_backlog()  # 1.23, Normally triggered by a timer
		self.assertBacklogs(location='flash', get=5, flash=5, storage=0)

		self.send_data_to_vrm(7.89) # sends 2.34 in the _send_data_from_backlog
		self.flash._send_data_from_backlog()  # 3.45, Normally triggered by a timer
		self.flash._send_data_from_backlog()  # 4.56, Normally triggered by a timer
		self.flash._send_data_from_backlog()  # 5.67, Normally triggered by a timer
		self.flash._send_data_from_backlog()  # 6.78, Normally triggered by a timer
		self.flash._send_data_from_backlog()  # 7.89, Normally triggered by a timer
		self.flash._send_data_from_backlog()  # update state, Normally triggered by a timer

		self.assertBacklogs(location='', get=0, flash=0, storage=0)
		self.assertEqual([0.12, 1.23, 2.34, 3.45, 4.56, 5.67, 6.78, 7.89], self._successfully_delivered)

		self.send_data_to_vrm(8.90)

		self.assertBacklogs(location='', get=0, flash=0, storage=0)
		self.assertEqual([0.12, 1.23, 2.34, 3.45, 4.56, 5.67, 6.78, 7.89, 8.90], self._successfully_delivered)

	def test_p06_resetting_vrmlogger_without_flash_backlog(self):
		'''The flash backlog should not be used on startup if it exists but it contains no data'''

		self.assertFalse(os.path.exists(self.tmp_flash_path))
		self.break_the_internet()

		self.send_data_to_vrm(0.12)
		self.send_data_to_vrm(1.23)
		self.send_data_to_vrm(2.34)
		self.assertBacklogs(location='flash', get=3, flash=3, storage=0)
		self.assertTrue(os.path.exists(self.tmp_flash_path))

		self.fix_the_internet()

		self.flash._send_data_from_backlog()  # Normally triggered by a timer
		self.flash._send_data_from_backlog()  # Normally triggered by a timer
		self.flash._send_data_from_backlog()  # Normally triggered by a timer
		self.flash._send_data_from_backlog()  # Normally triggered by a timer

		self.assertBacklogs(location='', get=0, flash=0, storage=0)

		logging.debug('Destroying existing Flash instance...')
		del self.flash

		logging.debug('Creating new Flash instance...')
		self.assertTrue(os.path.exists(self.tmp_flash_path))
		self.setupVrmHttpFlash()
		self.assertBacklogs(location='', get=0, flash=0, storage=0)

class VrmstorageBacklogTests(VrmHttpFlashTests):
	'''Test requirements and scenarios of VrmHttpFlash with logging to storage

	While not the case in real life, for the testcases it is assumed that the
	attribute value passed to the VRM Portal is strictly increasing over time.
	This way the requirement that all data should be logged chronologically can
	be tested by looking at the data attribute.
	'''

	def setUp(self, attribute_name='CV[0]', attribute_start_value=1234.5):
		self._logmode = 2
		VrmHttpFlashTests.setUp(self, attribute_name=attribute_name, attribute_start_value=attribute_start_value)
		self.fix_the_internet()

	def tearDown(self):
		self.assertEqual([], self._successfully_delivered)
		if (os.path.exists(LAST_KNOWN_STORAGEPATH_FILE)):
			os.unlink(LAST_KNOWN_STORAGEPATH_FILE)
		VrmHttpFlashTests.tearDown(self)

	def test_s01_empty_on_startup(self):
		self.assertBacklogs(location='', get=0, flash=0, storage=0)

	def test_s02_send_before_mount(self):
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.assertBacklogs(location='flash', get=4, flash=4, storage=0)

	def test_s03_send_after_mount(self):
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.assertBacklogs(location='flash', get=3, flash=3, storage=0)
		self.flash.mount_event('/dev/xxx', self.tmp_dir)
		self.assertBacklogs(location='flash', get=3, flash=3, storage=0)
		self.flash._store_data_from_backlog()
		self.assertBacklogs(location='', get=0, flash=0, storage=0)
		self.send_data_to_vrm()
		self.assertBacklogs(location='', get=0, flash=0, storage=1)

	def test_s04_startup_use_previous_location(self):
		self.flash.mount_event('/dev/xxx', self.tmp_dir)
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.assertBacklogs(location='', get=0, flash=0, storage=3)

		del self.flash

		logging.debug('Creating new Flash instance...')
		self.setupVrmHttpFlash()
		self.assertBacklogs(location='', get=0, flash=0, storage=0)
		self.send_data_to_vrm()
		self.assertBacklogs(location='', get=0, flash=0, storage=1)

	def test_s05_startup_with_previous_location_inaccesible(self):
		self.flash.mount_event('/dev/xxx', self.tmp_dir)
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.assertBacklogs(location='', get=0, flash=0, storage=3)

		del self.flash
		shutil.rmtree(self.tmp_dir)
		self.tmp_dir = tempfile.mkdtemp(suffix='vrmlogger_test')

		logging.debug('Creating new Flash instance...')
		self.setupVrmHttpFlash()
		self.assertBacklogs(location='', get=0, flash=0, storage=0)
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.assertBacklogs(location='flash', get=2, flash=2, storage=0)
		self.flash.mount_event('/dev/xxx', self.tmp_dir)
		self.flash._store_data_from_backlog()
		self.assertBacklogs(location='', get=0, flash=0, storage=0)
		self.assertEqual(1, self.flash._dbusservice['/Logscript/LogfilesOnStorage'])
		self.send_data_to_vrm()
		self.assertBacklogs(location='', get=0, flash=0, storage=1)
		self.assertEqual(2, self.flash._dbusservice['/Logscript/LogfilesOnStorage'])

	def test_s06_dbus_logfiles_on_storage(self):
		self.assertEqual(-1, self.flash._dbusservice['/Logscript/LogfilesOnStorage'])
		self.flash.mount_event('/dev/xxx', self.tmp_dir)
		self.assertEqual(0, self.flash._dbusservice['/Logscript/LogfilesOnStorage'])
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.assertBacklogs(location='', get=0, flash=0, storage=3)
		self.assertEqual(1, self.flash._dbusservice['/Logscript/LogfilesOnStorage'])

		del self.flash

		logging.debug('Creating new Flash instance...')
		self.setupVrmHttpFlash()
		self.assertBacklogs(location='', get=0, flash=0, storage=0)
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.assertBacklogs(location='', get=0, flash=0, storage=4)
		self.assertEqual(2, self.flash._dbusservice['/Logscript/LogfilesOnStorage'])

	def test_s07_inserting_into_flash(self):
		self.assertBacklogs(location='', get=0, flash=0, storage=0)
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.send_data_to_vrm()
		self.assertBacklogs(location='flash', get=5, flash=5, storage=0)

	def test_s08_moving_flash_backlog(self):
		self.send_data_to_vrm(0.12)
		self.send_data_to_vrm(1.23)
		self.send_data_to_vrm(2.34)
		self.send_data_to_vrm(3.45)
		self.send_data_to_vrm(4.56)
		self.send_data_to_vrm(5.67)
		self.send_data_to_vrm(6.78)
		self.assertBacklogs(location='flash', get=7, flash=7, storage=0)
		self.flash.mount_event('/dev/xxx', self.tmp_dir)
		self.flash._store_data_from_backlog()
		self.assertEqual(1, self.flash._dbusservice['/Logscript/LogfilesOnStorage'])
		self.assertBacklogs(location='', get=0, flash=0, storage=0)
		self.send_data_to_vrm(7.89)
		self.send_data_to_vrm(8.90)
		self.assertBacklogs(location='', get=0, flash=0, storage=2)
		self.assertEqual(2, self.flash._dbusservice['/Logscript/LogfilesOnStorage'])

	def test_s09_moving_flash_backlog_after_reset(self):
		self.send_data_to_vrm(0.12)
		self.send_data_to_vrm(1.23)
		self.send_data_to_vrm(2.34)
		self.send_data_to_vrm(3.45)
		self.send_data_to_vrm(4.56)
		self.send_data_to_vrm(5.67)
		self.send_data_to_vrm(6.78)
		self.assertBacklogs(location='flash', get=7, flash=7, storage=0)

		logging.debug('Destroying existing Flash instance...')
		del self.flash

		logging.debug('Creating new Flash instance...')
		self.setupVrmHttpFlash()
		# Because there is a previous location which is accesible and writable,
		# the sqlite3 database will be moved and renamed from the old location to the new
		# and the logger will start without backlog
		self.assertBacklogs(location='flash', get=7, flash=7, storage=0)
		self.flash.mount_event('/dev/xxx', self.tmp_dir)
		self.flash._store_data_from_backlog()
		self.assertEqual(1, self.flash._dbusservice['/Logscript/LogfilesOnStorage'])

		self.send_data_to_vrm(7.89)
		self.send_data_to_vrm(8.90)

		self.assertBacklogs(location='', get=0, flash=0, storage=2)
		self.assertEqual(2, self.flash._dbusservice['/Logscript/LogfilesOnStorage'])

if __name__ == "__main__":
	logging.basicConfig(stream=sys.stderr)
	#logging.getLogger('').setLevel(logging.WARNING)
	#logging.getLogger('').setLevel(logging.INFO)
	logging.getLogger().setLevel(logging.DEBUG)
	unittest.main()
