#!/usr/bin/python -u
# -*- coding: utf-8 -*-

## @package dbus_vrm
# Run this with:
#	export PYTHONPATH="../velib_python"; python vrmlogger.py -d

# TODO: make some script that puts values on dbus which makes it easy to test the different whentolog options
# And see more todo's in code.

from dbus.mainloop.glib import DBusGMainLoop
import gobject
import argparse
import logging
import datetime
import platform
import dbus
import os
import signal
import sys
import time
from random import randint

logger = logging.getLogger(__name__)

# Victron imports
sys.path.insert(1, os.path.join(os.path.dirname(__file__), './ext/velib_python'))
import datalist
from dbusmonitor import DbusMonitor
from vedbus import VeDbusService
from vedbus import VeDbusItemImport
from vrmhttp import VrmCommandType, VrmHTTP, vrm_encode
from http_endpoint import HttpEndpoint
from storage_endpoint import StorageEndpoint
from settingsdevice import SettingsDevice
from conversions import Conversions
from kwhdeltas import KwhDeltas
from networktrafficdeltas import NetworkTrafficDeltas
from nonvolatilequeue import NonVolatileQueue
from ve_utils import exit_on_error

softwareversion = '2.34'

# list, containing the VeDbusImportItem objects of values that are only logged on change, and that have
# changed since the last log. It is emptied everytime the values are sent.
changedValues = {}

# refers to the a timer that is started when a dbus value changes that triggers a configchange. Often
# there are several of such value changes that follow each other. Therefore don't send a configmessage
# on the first change, but wait a few more seconds to get the rest in as well.
sendconfigdelaytimer = None

# contains the gobject timer source id for the hourly timer
hourlytimer = None

# contains the gobject timer source id for the interval timer
senddataintervaltimer = None

# same as above, but then for the onchangeevent type data
senddatadelaytimer = None

# our own representation on the dbus
dbusservice = None

# this is where all the data is going
endpoint = None

# link to com.victronenergy.settings, where our config is stored
settings = None

# KwhDeltas instance
kwhdeltas = None

networktrafficdeltas = None

mainloop = None

buffer = None

class LOGMODE:
	LoggerDisabled, LogToVRM, LogToStorage = range(3)
	names = {0: 'LoggerDisabled', 1: 'LogToVRM', 2: 'LogToStorage'}

def get_announce():
	try:
		# one dir up from the dir where this script is, we expect a file called version
		fileObject = open(os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir, 'version')), 'r')
		lines = fileObject.readlines()
		version = lines[0].rstrip('\n')
		build = lines[2].rstrip('\n')
		fileObject.close()
	except Exception, ex:
		logger.error('get_announce() => exception while reading ccgx-version information: %s. We expect file ../version' % ex)
		version = '?'
		build = '?'

	try:
		fileObject = open('/tmp/last_boot_type', 'r+')
		line = fileObject.readline()
		boot_type = line.rstrip('\n')
		if boot_type != -3:
			fileObject.seek(0,0)
			fileObject.write('-3\n')  # Write to the file that we have already read it.
		fileObject.close()
	except Exception, ex:
		logger.error('get_announce() => exception while opening /tmp/last_boot_type: %s.' % ex)
		boot_type = '-2'

	return {'v': version, 'build': build, 'w': boot_type}


def get_configchange():
	a = dbusMonitor.get_values(['configChange'], Conversions())

	# Get the ipaddesses, but don't make ourselves dependent on connman, hence the try/except.
	addresses = ""
	try:
		manager = dbus.Interface(dbusservice.dbusconn.get_object("net.connman", "/"), "net.connman.Manager")
		for path, properties in manager.GetServices():
			if "IPv4" in properties:
				ipv4 = properties["IPv4"]
				if "Address" in ipv4:
					addresses += ipv4["Address"] + ";"
			if "IPv6" in properties:
				ipv6 = properties["IPv6"]
				if "Address" in ipv6:
					addresses += ipv6["Address"] + ";"
	except Exception, ex:
		logger.error("Failed to get local ip addresses; %s" % ex)

	a['ip'] = addresses.rstrip(";")

	return a


#	dbus-servicename, for example com.victronenergy.dbus.ttyO1
#	dbus-path, for example /Ac/ActiveIn/L1/V
#   options: the dict containing the properties from the vrmTree
#	changes: the changes, a tuple with GetText() and GetValue()
#	deviceInstance: the device instance.
def value_changed_on_dbus(dbusServiceName, dbusPath, options, changes, deviceInstance):
	# logger.debug('sendconfigdelaytimer: %s, senddatadelaytimer: %s' % (sendconfigdelaytimer, senddatadelaytimer))
	# logger.debug(
	#	"%s%s: options: %s, changes: %s, instance: %s" %
	#	(dbusServiceName, dbusPath, options, changes, deviceInstance))

	if options['whenToLog'] == 'configChange':
		initiate_send_config()

	elif options['whenToLog'] == 'onIntervalAlwaysAndOnEvent':
		initiate_send_data()

	elif options['whenToLog'] == 'onIntervalOnlyWhenChanged':
		global changedValues
		# TODO regarding below None exception: does it make sense to send an empty string when a value is
		# being invalidated, instead of just sending nothing? We cannot guarantee that sending an empty
		# string is always sent. CCGX could be switched of at an unfortunate time for example. Or a device
		# is disconnected completely. One reason why it would make sense is to deleted it from the
		# cache in table lastLogData (?) while the rest of the system keeps running and sending data.
		# Anyway, at this moment the vrm db api can probably not cope with empty strings (?), and sending an
		# empty string will not trigger a delete from lastLogData. Therefore for now just hide it, to prevent
		# raising an error (unsupported operand type(s) for /: 'NoneType' and 'int')

		# Most common situation for values, type onIntervalOnlyWhenChanged, coming in as None, is when a
		# device is being disconnected: the D-Bus service then invalidates all values before going offline.
		if changes['Value'] is not None:
			changedValues[options['code'] + '[' + str(deviceInstance) + ']'] = Conversions.convert(
				options['code'], changes['Value'])


# Called from the dbusmonitor
def handle_new_service(servicename, deviceinstance):
	kwhdeltas.handle_new_service(servicename, deviceinstance)
	initiate_send_config()

	# Add all values that are normally only sent on change
	global changedValues
	changedValues = dbusMonitor.get_values(['onIntervalOnlyWhenChanged'], Conversions())

	initiate_send_data()

# Called from the dbusmonitor
def handle_removed_service(servicename, deviceinstance):
	kwhdeltas.handle_removed_service(servicename, deviceinstance)
	initiate_send_config()


def initiate_send_config():
	global sendconfigdelaytimer
	if sendconfigdelaytimer is None:
		logger.debug('Initiating a config change, starting the configchangedelaytimer (5 seconds)')
		sendconfigdelaytimer = gobject.timeout_add(5000, exit_on_error, send_config)


def initiate_send_data():
	global senddatadelaytimer
	if senddatadelaytimer is None:
		logger.debug('Initiating a send-data, starting the senddatadelaytimer (5 seconds)')
		senddatadelaytimer = gobject.timeout_add(5000, exit_on_error, send_data)


def send_config():
	global sendconfigdelaytimer

	if sendconfigdelaytimer:
		gobject.source_remove(sendconfigdelaytimer)
		sendconfigdelaytimer = None

	endpoint.send(VrmCommandType.CONFIGCHANGE, get_configchange())

	return False  # stop the timer calling this function, only timer calling this function is sendconfigdelaytimer


def send_data():
	global senddataintervaltimer
	global changedValues
	global senddatadelaytimer

	if senddataintervaltimer:
		gobject.source_remove(senddataintervaltimer)

	if settings['logmode'] > 0:
		senddataintervaltimer = gobject.timeout_add(
			settings['interval'] * 1000, exit_on_error, send_data)

	if senddatadelaytimer:
		gobject.source_remove(senddatadelaytimer)
		senddatadelaytimer = None

	data = dbusMonitor.get_values(['onIntervalAlwaysAndOnEvent', 'onIntervalAlways'], Conversions())
	data.update(changedValues)
	changedValues = {}

	data.update(kwhdeltas.getdeltas(refreshbaseline=False))

	endpoint.send(VrmCommandType.SENDDATA, data)

	# stop the timer calling this function, could be de senddatadelaytimer or the senddataintervaltimer:
	# If it is the senddatadelaytimer, obviously it shouldnt trigger another send in 5 seconds.
	# If it was the senddataintervaltimer, cancel it as well: it is restarted anyway, see start of this
	# function.
	return False


def send_hourlydeltas():
	global hourlytimer

	data = kwhdeltas.getdeltas(refreshbaseline=True)
	data.update(networktrafficdeltas.getdeltas(refreshbaseline=True))

	# Since all gateways out there will send their hourly update, randomly delay sending the data a
	# litle, so that the load on the ramshackle vrm-backend server is distributed, a little. :-)
	# The lamda is used to make sure that the time_out is stopped after one call
	gobject.timeout_add(
		randint(0, 30000), exit_on_error,
		lambda: endpoint.send(VrmCommandType.HOURLYDELTAS, data) and False)

	gobject.source_remove(hourlytimer)
	hourlytimer = gobject.timeout_add(secondsuntilnexthour() * 1000, exit_on_error, send_hourlydeltas)


def gettext_fromtimestamp(path, value):
	return datetime.datetime.fromtimestamp(int(value)).strftime('%Y-%m-%d %H:%M:%S')


def flush_back_log(path, value):
	logger.warning('No longer implemented')

	# By returning False we prevent the value from changing, allowing it to be set over and over.
	return False


def handle_changed_setting(setting, oldvalue, newvalue):
	global endpoint
	global senddataintervaltimer

	if setting == 'logmode' and newvalue == 0:
		logger.debug("Someone disabled the logger: stop")
		stop_logging()

	elif setting == 'logmode' and oldvalue == 0:
		logger.debug("Someone enabled the logger: start")
		start_logging()

	elif setting == 'logmode':
		logger.debug("Someone changed the logmode: %s -> %s" % (oldvalue, newvalue))
		stop_logging()
		start_logging(False)

	elif setting == 'interval':
		stop_logging()
		start_logging(False)

	elif setting == 'url' or setting == 'proxyaddress' or setting == 'proxyport':
		if settings['logmode'] == 1:
			stop_logging()
			start_logging(False)

	else:
		raise Exception("ERROR: SOME SETTINGS CONDITION CHANGED THAT WE DON'T KNOW? (%s, %s)" % (setting, newvalue))


def stop_logging():
	global senddataintervaltimer
	assert senddataintervaltimer is not None, \
		"Stopping to log, so asserted that senddataintervaltimer is not None"
	gobject.source_remove(senddataintervaltimer)
	senddataintervaltimer = None

	global hourlytimer
	assert hourlytimer is not None, \
		"Stopping to log, se asserted that hourlytimer is not None"
	gobject.source_remove(hourlytimer)
	hourlytimer = None

	global endpoint
	if hasattr(endpoint, 'stop'):
		endpoint.stop()  # blocking, it waits until the vrm worker is stopped. max time is same
						  # as vrm http timeout
	del endpoint

	logger.info("logger stopped")


def start_logging(doinitialsend=True):
	global senddataintervaltimer
	assert senddataintervaltimer is None, \
		"Starting to log, so asserted that senddataintervaltimer is None"

	url = settings['url']
	if url == '':
		url = 'http://ccgxlogging.victronenergy.com/log/log.php'

	logger.info("Starting to log, logmode == %s" % LOGMODE.names[settings['logmode']])

	global endpoint
	if settings['logmode'] == LOGMODE.LogToStorage:
		endpoint = StorageEndpoint(
			interval=settings['interval'],
			buffer=buffer,
			dbusconn=dbusMonitor.dbusConn,
			update_storage_count=lambda c: dbusservice.__setitem__('/Storage/CurrentFile/Count', c),
			update_logfiles_count=lambda c: dbusservice.__setitem__('/Storage/LogFilesCount', c),
			update_mount_state=lambda c: dbusservice.__setitem__('/Storage/MountState', c))
	elif settings['logmode'] == LOGMODE.LogToVRM:
		endpoint = HttpEndpoint(
			interval=settings['interval'],
			buffer=buffer,
			logurl=url,
			update_last_contact=lambda t: dbusservice.__setitem__('/Vrm/TimeLastContact', t),
			)

	if doinitialsend:
		# we booted up, send the announce message.
		endpoint.send(VrmCommandType.ANNOUNCE, get_announce())

		# send a configchange
		send_config()

		# Add all values that are normally only sent on change
		global changedValues
		changedValues = dbusMonitor.get_values(['onIntervalOnlyWhenChanged'], Conversions())

	assert senddataintervaltimer is None, \
		"Starting to log, asserted that senddataintervaltimer is None"

	# Send data, which will automatically start the interval timer
	send_data()

	global hourlytimer
	assert hourlytimer is None, \
		"Starting to log, asserted that hourlytimer is None"
	hourlytimer = gobject.timeout_add(secondsuntilnexthour() * 1000, exit_on_error, send_hourlydeltas)

def on_mount_state_changed(path, new_value):
	logger.info('{} - {}'.format(path, new_value))
	assert path == '/Storage/MountState'
	if new_value == 2:
		global endpoint
		# Ugly: from an object oriented point of view, we should add an eject function to all endpoint
		# classes, but it is storage specific, so we swallow the AttributeError instead.
		try:
			endpoint.eject()
		except AttributeError:
			pass

def secondsuntilnexthour():
	return 3600 - (int(time.time()) % 3600)

def sig_handler(signum, frame):
	global mainloop
	mainloop.quit()

def main():
	global dbusMonitor
	global dbusservice
	global settings
	global mainloop

	gobject.threads_init()

	# Argument parsing
	parser = argparse.ArgumentParser(
		description='vrmlogger v%s: communication to VRM Portal database' % softwareversion
	)

	parser.add_argument("-d", "--debug", help="set logging level to debug",
					action="store_true")

	args = parser.parse_args()

	# Init logging
	FORMAT = '%(threadName)s-%(module)s: %(message)s'

	logging.basicConfig(level=(logging.DEBUG if args.debug else logging.INFO), format=FORMAT)
	logging.info("%s v%s is starting up" % (__file__, softwareversion))
	logLevel = {0: 'NOTSET', 10: 'DEBUG', 20: 'INFO', 30: 'WARNING', 40: 'ERROR'}
	logging.info('Loglevel set to ' + logLevel[logging.getLogger().getEffectiveLevel()])

	# Have a mainloop, so we can send/receive asynchronous calls to and from dbus
	DBusGMainLoop(set_as_default=True)

	# Put ourselves onto the dbus as a service
	dbusservice = VeDbusService("com.victronenergy.logger")
	dbusservice.add_path('/Vrm/TimeLastContact', value=None,
							description="Unixtimestamp of last successfull contact with VRM DB API",
							writeable=False, gettextcallback=gettext_fromtimestamp)

	dbusservice.add_path('/Buffer/Count', value=None,
							description="The number of entries in the buffer on the local flashdisk",
							writeable=False)

	dbusservice.add_path('/Buffer/OldestTimestamp', value=None,
							description="Unixtimestamp of the oldest item pending in the buffer",
							writeable=False, gettextcallback=gettext_fromtimestamp)

	dbusservice.add_path('/Storage/CurrentFile/Count', value=None,
							description="The number of entries in the current external storage",
							writeable=False)

	dbusservice.add_path('/Storage/LogFilesCount', value=None,
							description="The amount of sqlite files on the external storage",
							writeable=False)

	dbusservice.add_path('/Storage/MountState', value=0,
							description="State of external storage device (0 not mounted, 1 mounted, 2 unmount desired, 3 busy unmounting)",
							writeable=True, onchangecallback=on_mount_state_changed)

	global kwhdeltas
	kwhdeltas = KwhDeltas()

	global networktrafficdeltas
	networktrafficdeltas = NetworkTrafficDeltas()

	# add the dbus-data that kwhdeltas need to datalist.vrmtree without overwriting it
	kwhtree = kwhdeltas.get_dbusmonitortree()
	for serviceclass, newpaths in kwhtree.items():
		if serviceclass in datalist.vrmtree:
			# the class is already in there, check if all paths are as well.
			existingpaths = datalist.vrmtree[serviceclass]
			for newpath, newoptions in newpaths.items():
				if newpath not in existingpaths:
					existingpaths[newpath] = newoptions
		else:
			# class is not yet there, simply add it
			datalist.vrmtree[serviceclass] = newpaths

	dbusMonitor = DbusMonitor(
		datalist.vrmtree,
		value_changed_on_dbus,
		deviceAddedCallback=handle_new_service,
		deviceRemovedCallback=handle_removed_service,
	)

	kwhdeltas.start(dbusMonitor)

	bus = dbus.SystemBus() if (platform.machine() == 'armv7l') else dbus.SessionBus()
	serviceName = 'com.victronenergy.settings'

	# Try to read the log mode item. Do this before creating the SettingsDevice, because that one will add
	# the path when it does not exist yet.
	logModeItem = VeDbusItemImport(bus, serviceName, path='/Settings/Vrmlogger/Logmode', createsignal=False)
	firstTimeNewSettings = not logModeItem.exists

	# Hm... What happens here when the Logmode item does not exist. It is created here and do we get a changed callback then??
	settings = SettingsDevice(
		bus,
		supportedSettings={
			'logmode': ['/Settings/Vrmlogger/Logmode', 1, 0, 2],
			'proxyaddress': ['/Settings/Vrmlogger/Http/Proxy', '', 0, 0],
			'proxyport': ['/Settings/Vrmlogger/Http/ProxyPort', '', 0, 0],
			'interval': ['/Settings/Vrmlogger/LogInterval', 900, 0, 0],
			'url': ['/Settings/Vrmlogger/Url', '', 0, 0]  # When empty, the default url will be used.
			},
		eventCallback=handle_changed_setting)

	if firstTimeNewSettings == True:
		logger.info("Converting old settings (/Settings/Logscript/*) to new settings (/Settings/Vrmlogger/*)")

		# Read all the old items
		oldItem = VeDbusItemImport(bus, serviceName, path='/Settings/Logscript/Enabled', createsignal=False)
		value = oldItem.get_value()
		if value is not None:
			settings['logmode'] = value

		oldItem = VeDbusItemImport(bus, serviceName, path='/Settings/Logscript/Http/Proxy',	createsignal=False)
		value = oldItem.get_value()
		if value is not None:
			settings['proxyaddress'] = value

		oldItem = VeDbusItemImport(bus, serviceName, path='/Settings/Logscript/Http/ProxyPort',	createsignal=False)
		value = oldItem.get_value()
		if value is not None:
			settings['proxyport'] = value

		oldItem = VeDbusItemImport(bus, serviceName, path='/Settings/Logscript/LogInterval', createsignal=False)
		value = oldItem.get_value()
		if value is not None:
			settings['interval'] = value

		oldItem = VeDbusItemImport(bus, serviceName, path='/Settings/Logscript/Url', createsignal=False)
		value = oldItem.get_value()
		if value is not None:
			settings['url'] = value

	# Replace the old default url with a better default value: empty string
	if settings['url'] == 'http://vrm.victronenergy.com/log/log.php':
		settings['url'] = ''

	# TODO: look in the really old vrmlogger (Arnouds version) to find out where the backlog was stored
	# and delete that file
	global buffer
	buffer = NonVolatileQueue(
		max_age=3600 * 24 * 2,  # Two days
		path='/data/db/vrmlogger-backlog.sqlite3',
		update_count=lambda c: dbusservice.__setitem__('/Buffer/Count', c),
		update_oldest_timestamp=lambda t: dbusservice.__setitem__('/Buffer/OldestTimestamp', t))

	logger.info("Backlog buffer initialised. Count == %d" % buffer.count)

	if settings['logmode'] > 0:
		start_logging()
	else:
		logger.info('Init finished, but not starting: logmode == %s' % LOGMODE.names[settings['logmode']])

	# Start and run the mainloop
	logger.info("Starting mainloop, responding on only events")
	mainloop = gobject.MainLoop()
	signal.signal(signal.SIGTERM, sig_handler)
	signal.signal(signal.SIGINT, sig_handler)
	mainloop.run()
	if settings['logmode'] > 0:
		stop_logging()

if __name__ == "__main__":
	main()
