# Support for remote vebus device settings and configuration (VEConfigure/VE.Bus System Configurator)

import subprocess
#import ftplib
import os
import time
#import re
import errno
import logging
logger = logging.getLogger(__name__)

from dbus.mainloop.glib import DBusGMainLoop
from dbusmonitor import DbusMonitor
from vedbus import VeDbusService
from vedbus import VeDbusItemImport

from shared import get_vrm_portal_id

MK2VSC_PROGRAM = "/opt/color-control/mk2vsc/mk2vsc"

mk2VscDbusMon = None
mk2VscDBusStatusMonitor = None

mk2VscTree = {
'com.victronenergy.mk2vsc': {
	'/State': {'code': 'S', 'whenToLog': 'onIntervalAlwaysAndOnEvent', 'accessLevel': ''}
}}

mk2VscState = {
	# mk2vsc application state codes
	10 : "Init",
	11 : "Query products",
	12 : "Done",
	21 : "Read setting data",
	22 : "Read assistants",
	23 : "Read VEBus configuration",
	30 : "Write settings info",
	31 : "Write Settings Data",
	32 : "Write assistants",
	33 : "Write VEBus configuration",
	40 : "Resetting VEBus products"
}

class Mk2VscError(Exception):
	def __init__(self, errno):
		self.errors = {
			# mk2vsc application error/return codes
			30 : "MK2 communication error",
			31 : "Product address not reachable",
			32 : "Incompatible MK2 firmware version",
			33 : "No VEBus product was found",
			34 : "Too many devices on the VEBus",
			35 : "Timed out",
			36 : "Wrong password",
			40 : "Malloc error",
			50 : "File open error",
			51 : "File write error",
			52 : "File read error",
			53 : "File checksum error",
			54 : "File incompatible version number",
			55 : "File section not found",
			56 : "File format error",
			57 : "Product target number does not match file",
			58 : "File expired",
			60 : "VEBus configuration write failed",
			61 : "Incompatible VEBus configuration",
			70 : "Assistants write failed",
			71 : "Assistants read failed",

			# vrmmk2vsc python module error codes
			100 : "OSerror, unknown application",
			# Velib error code. (offset = 200)
			201 : "Failed to open com port(no response)"
		}
		if errno < 30:
			# mk2vsc can exit with velib error codes.
			# For instance error code 1 when open com-port fails.
			# When these errors occur velib causes the application to exit without returning to the main application.
			# Since we allready registered error code 1 in vrmpubnub we remap the codes to > 200
			errno = errno + 200
		self.errno = errno
	def __str__(self):
		return repr(self.errors.get(self.errno,"Unknown error"))


def create_filename(myId, mk2addr): # Without file extension
	inst = mk2addr.replace('dbus://com.victronenergy.vebus.','')
	inst = inst.replace('/', '_')
	return str(myId) + "_inst_" + inst


def list_products_addr(mk2Addr):
	productAddrList = []
	logger.info ("list products")
	try:
		mk2vscOutput = subprocess.check_output([MK2VSC_PROGRAM, "-l", "-s" ,mk2Addr])
		productListing = mk2vscOutput.splitlines()

		if "Connected products:" in productListing[0]:
			productListing.pop(0) # remove header
			for product in productListing:
				for s in product.split():
					if s.isdigit():
						productAddrList.append(int(s))
	except OSError as e:
		logger.info("mk2vsc OS error [" + str(e.errno) + "] " + e.strerror)
		raise Mk2VscError(100)
	except subprocess.CalledProcessError as e:
		logger.info("mk2vsc application error = " + str(e.returncode))
		raise Mk2VscError(e.returncode)
		# No products connected keep empty list
	return productAddrList


def read_settings(myId, mk2Addr, vscFileDir):
	fileName = create_filename (myId, mk2Addr)
	dirAndFileName = vscFileDir + fileName

	# Remove any existing file with the same name
	files = os.listdir(vscFileDir)
	for file in files:
		if (file == fileName + ".rvsc") or (file == fileName + ".rvms"):
			try:
				os.remove(vscFileDir + file )
			except OSError:
				pass

	logger.info("read_settings")

	try:
		subprocess.check_call([MK2VSC_PROGRAM, "-r",
						 "-f", dirAndFileName, "-s" ,mk2Addr])
	except OSError as e:
		logger.info("mk2vsc OS error [" + str(e.errno) + "] " + e.strerror)
		raise Mk2VscError(100)
	except subprocess.CalledProcessError as e:
		print("mk2vsc error = " + str(e.returncode))
		raise Mk2VscError(e.returncode)

	# Find file and add the correct extention
	if os.path.exists(vscFileDir + fileName + ".rvsc"):
		return fileName + ".rvsc"
	if os.path.exists(vscFileDir + fileName + ".rvms"):
		return fileName + ".rvms"

	return None


def write_settings(myId, mk2Addr, vscFileDir, vscFileName):
	logger.info("write_settings from file %s" % vscFileName)

	try:
		subprocess.check_call([MK2VSC_PROGRAM, "-w",
						 "-f", vscFileName, "-s" ,mk2Addr])
		# vscFileName includes the directory
	except OSError as e:
		logger.info("mk2vsc OS error [" + str(e.errno) + "] " + e.strerror)
		raise Mk2VscError(100)
	except subprocess.CalledProcessError as e:
		logger.info("mk2vsc application error = " + str(e.returncode))
		raise Mk2VscError(e.returncode)

	return os.path.basename(vscFileName)


def value_changed_on_dbus(dbusServiceName, dbusPath, dict, changes, deviceInstance):
	statuscode = changes['Value']
	if mk2VscDBusStatusMonitor != None:
		mk2VscDBusStatusMonitor(statuscode, mk2VscState.get(statuscode,"Unknown state"))


class Mk2Vsc:
	def __init__(self,statusMonitor, vscFileDir):
		self.myId = get_vrm_portal_id()
		self.vscFileDir = vscFileDir

		# Monitor the mk2vsc program state
		# We want to register the dbusmonitor only once.
		#TODO find a way to properly remove our tree from the debusmonitor
		global mk2VscDbusMon
		global mk2VscDBusStatusMonitor

		mk2VscDBusStatusMonitor = statusMonitor

		if mk2VscDbusMon == None:
			mk2VscDbusMon = DbusMonitor(mk2VscTree, value_changed_on_dbus,
										deviceAddedCallback=None, deviceRemovedCallback=None)

		# We need a vsc file data direcory. Create it in case it is not already available
		try:
			os.makedirs(self.vscFileDir) # try to create it
		except OSError as e:
			if e.errno != errno.EEXIST: # skip error
				raise

	def __enter__(self):
		return self

	def __exit__(self, type, value, traceback):
		return

	def list(self, mk2Addr):
		listing = list_products_addr(mk2Addr)
		return listing

	def read(self, mk2Addr):
		return read_settings(self.myId, mk2Addr, self.vscFileDir)

	def write(self, mk2Addr, vscFileName):
		return write_settings(self.myId, mk2Addr, self.vscFileDir, vscFileName)


def main():
	#test = Mk2Vsc()
	#test.list()

	with Mk2Vsc() as vsc:
		print("test code")
		try:
			#vsc.list(0)
			vsc.read(0,10)
		except Mk2VscError as e:
			print("mk2vsc error = " + str(e))
			print("mk2vsc errorcode = " + str(e.errno))









if __name__ == "__main__":
	main()



#def respond_to_svc_read(incomingchannel, parameters):
#def respond_to_svc_write(incomingchannel, parameters):
#def respond_to_svc_list(incomingchannel, parameters):
